import numpy as np


def compute_ranks(values):
    i = 1
    r = 1
    ranks = [1]

    while i < len(values):
        r = r + 1  # increase rank by one (create new rank)
        if values[i] != values[i - 1]:
            ranks.append(r)  # give new rank
        else:
            ranks.append(ranks[i - 1])  # give the same rank as before

        i = i + 1

    if len(ranks) == len(values):
        return ranks
    else:
        print("number of ranks does not equal number of values")


def compute_rank_p_vals(ranks):
    ranks_size = len(ranks)
    return [rank / ranks_size for rank in ranks]


def compute_log_10_p_vals(pvals):
    log_pvals = [np.log10(pval) * -1 for pval in pvals]
    return np.round(log_pvals, 3)


def rank_across_genome(test_data, top_lowest):

    # The test files always have the test values saved in the last column
    test_name = test_data.columns[-1]

    # get rid of results with nan (there should't be too many of them)
    test_data.dropna(axis=0, inplace=True)

    # Sort the data, so that the ranks can be given
    test_data.sort_values(by=test_name, inplace=True, ascending=top_lowest)

    test_results = test_data[test_name].values

    ranks = compute_ranks(test_results)
    test_data["rank"] = ranks

    rank_p_vals = compute_rank_p_vals(ranks)
    test_data["rank_p_value"] = rank_p_vals

    log_10_p_vals = compute_log_10_p_vals(rank_p_vals)
    test_data["-log10_p_value"] = log_10_p_vals

    return test_data



# Expose main modules
from .prepare import prepare
from .compute import compute
from .plot import plot

# Expose utility modules
from . import xp_utils
from . import utils
from . import rank_tools

__version__ = "1.1.4"



import sys
import click
from exp_heatmap import prepare, compute, plot, __version__

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

@click.group(help="ExP Heatmap - Population genetics visualization tool", epilog="For more information, see the documentation at:\nhttps://github.com/bioinfocz/exp_heatmap/", context_settings=CONTEXT_SETTINGS)
@click.version_option(__version__, '-v', '--version', prog_name='exp_heatmap')
def cli():
    pass

# prepare command
@cli.command(name='prepare', short_help='Convert VCF file to ZARR format', context_settings=CONTEXT_SETTINGS)
@click.argument('vcf_file', type=click.Path(exists=True, readable=True, dir_okay=False), required=True, metavar='<vcf_file>')
@click.option('-o', '--output', type=click.Path(), default='zarr_output', show_default=True, help='Directory for ZARR files')
def prepare_cmd(vcf_file, output):
    """
    <vcf_file>  PATH  Recoded VCF file
    """
    prepare(vcf_file, output)

# compute command
@cli.command(name='compute', short_help='Compute population genetics statistics', context_settings=CONTEXT_SETTINGS)
@click.argument('zarr_dir', type=click.Path(exists=True, readable=True, file_okay=False), required=True, metavar='<zarr_dir>')
@click.argument('panel_file', type=click.Path(exists=True, readable=True, dir_okay=False), required=True, metavar='<panel_file>')
@click.option('-o', '--output', type=click.Path(), default='output', show_default=True, help='Directory for output files')
@click.option('-t', '--test', type=click.Choice(['xpehh', 'xpnsl', 'delta_tajima_d', 'hudson_fst']), default='xpehh', show_default=True, help='Statistical test to compute')
@click.option('-c', '--chunked', is_flag=True, help='Use chunked array to avoid memory exhaustion')
def compute_cmd(zarr_dir, panel_file, output, test, chunked):
    """
    <zarr_dir>  PATH  Directory with ZARR files from 'exp_heatmap prepare'
    <panel_file>  PATH  Population panel file
    """
    compute(zarr_dir, panel_file, output, test, chunked)

# plot command
@cli.command(name='plot', short_help='Generate heatmap visualization', context_settings=CONTEXT_SETTINGS)
@click.argument('input_dir', type=click.Path(exists=True, file_okay=False), required=True, metavar='<input_dir>')
@click.option('-s', '--start', type=int, help='Start position for the displayed region.')
@click.option('-e', '--end', type=int, help='End position for the displayed region.')
@click.option('-m', '--mid', type=int, help='Middle of the displayed area. The start and end positions will be calculated (mid ± 500 kb)')
@click.option('-t', '--title', type=str, help='Title of the heatmap')
@click.option('-o', '--output', type=click.Path(), default='ExP_heatmap',show_default=True, help='The output heatmap')
@click.option('-c', '--cmap', type=str, default='Blues', show_default=True, help='Matplotlib colormap for heatmap visualization')
def plot_cmd(input_dir, start, end, mid, title, output, cmap):
    """
    <input_dir>  PATH  Directory with TSV files from 'exp_heatmap compute'
    
    For positional arguments, use either [-s with -e] or [-m].
    """
    start_end_provided = start is not None and end is not None
    mid_provided = mid is not None
    
    #Check for invalid combinations of arguments
    if (start is None) != (end is None):
        raise click.UsageError("--start and --end must be used together")
    if not start_end_provided and not mid_provided:
        raise click.UsageError("Either (--start and --end) or --mid must be provided")
    if start_end_provided and mid_provided:
        raise click.UsageError("Cannot use both (--start and --end) and --mid at the same time")
    
    start_pos, end_pos = (mid - 500000, mid + 500000) if mid_provided else (start, end)
    plot(input_dir, start=start_pos, end=end_pos, title=title, output=output, cmap=cmap)

if __name__ == "__main__":
    cli()


import os
import sys
import seaborn as sns
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import glob
from bisect import bisect_left
from tqdm import tqdm

# 1000 Genomes Project data
populations_1000genomes = ("ACB","ASW","ESN","GWD","LWK","MSL","YRI","BEB","GIH","ITU","PJL","STU","CDX","CHB","CHS","JPT","KHV","CEU","FIN","GBR","IBS","TSI","CLM","MXL","PEL","PUR")
superpopulations = {"AFR": ("ACB", "ASW", "ESN", "GWD", "LWK", "MSL", "YRI"),
                    "SAS": ("BEB", "GIH", "ITU", "PJL", "STU"),
                    "EAS": ("CDX", "CHB", "CHS", "JPT", "KHV"),
                    "EUR": ("CEU", "FIN", "GBR", "IBS", "TSI"),
                    "AMR": ("CLM", "MXL", "PEL", "PUR",)}

def create_plot_input(input_dir, start, end, populations="1000Genomes", rank_pvalues="2-tailed"):
    """
    Generate a pandas DataFrame for plotting from a directory of pairwise population .tsv files.

    Parameters:
        input_dir (str): Directory containing .tsv files named as 'POP1_POP2.*.tsv' for each population pair.
        start (int): Start genomic position (X-axis lower bound).
        end (int): End genomic position (X-axis upper bound).
        populations (iterable or "1000Genomes"): List of population names to include and order in the heatmap. If not using 1000 Genomes, provide a custom iterable.
        rank_pvalues (str): Determines which p-value column to use:
            - "ascending": Use "-log10_p_value_ascending".
            - "descending": Use "-log10_p_value_descending" (default).
            - "2-tailed": For pop1_pop2, use ascending; for pop2_pop1, use descending.

    Returns:
        pd.DataFrame: Formatted data for heatmap plotting.
    """
    
    df_list = []
    pop_id_list = []
    population_sorter = populations_1000genomes if populations=="1000Genomes" else populations
    segment_files = glob.glob(os.path.join(input_dir, "*.tsv"))
    index = 1
    
    # Load population pair files between start and end and check if the populations are in the population_sorter
    for segment_file in tqdm(segment_files, desc="Loading population pair files"):
        pop_pair = os.path.splitext(os.path.basename(segment_file))[0].split(".")[0]
        p1, p2 = pop_pair.split("_")
        
        if all(x in population_sorter for x in (p1, p2)):
            pop_id_list.append(pop_pair)

            segments = pd.read_csv(segment_file, sep="\t")
            segments = segments[(segments.variant_pos >= start) & (segments.variant_pos <= end)]

            df_list.append(segments)

        else:
            print(f"[{index}/{len(segment_files)}] ERROR Loading {pop_pair} from {segment_file}. {p1} or {p2} not in provided population list.")
            
        index += 1
            
    # Validate the dimensions and variant_pos in each dataframe
    df_shape = df_list[0].shape
    variant_positions = df_list[0].variant_pos.values

    for i in range(len(df_list)):
        if df_list[i].shape != df_shape:
            raise ValueError(f"the shapes dont match in df {i}")
        if not np.array_equal(df_list[i].variant_pos.values, variant_positions):
            raise ValueError(f"the variant_positions dont match in df {i}")

    # Select variant_pos and -log10_p_value and transpose each df
    transp_list = []

    for df, pop_pair in zip(df_list, pop_id_list):
        if rank_pvalues == "ascending":
            pvalues_column1 = "-log10_p_value_ascending"
            pvalues_column2 = "-log10_p_value_ascending"
        
        elif rank_pvalues == "descending":
            pvalues_column1 = "-log10_p_value_descending"
            pvalues_column2 = "-log10_p_value_descending"

        elif rank_pvalues == "2-tailed":
            pvalues_column1 = "-log10_p_value_descending"
            pvalues_column2 = "-log10_p_value_ascending"
        
        else:
            raise ValueError(f"Unknown value for 'rank_pvalues' parameter in create_plot_input(). Expected values are: 'ascending', 'descending' or '2-tailed', got '{rank_pvalues}'")
            
        # Extract and transpose p-values for pop1_pop2 (pop1 under selection)
        left_df = df[["variant_pos", pvalues_column1]].copy()
        left_df.rename(columns={pvalues_column1: pop_pair}, inplace=True)
        left_df = left_df.set_index("variant_pos").T
        transp_list.append(left_df)

        # Extract and transpose p-values for pop2_pop1 (pop2 under selection)
        reverse_pop_pair = "_".join(pop_pair.split("_")[::-1])
        right_df = df[["variant_pos", pvalues_column2]].copy()
        right_df.rename(columns={pvalues_column2: reverse_pop_pair}, inplace=True)
        right_df = right_df.set_index("variant_pos").T
        transp_list.append(right_df)

    # Concatenate all transposed DataFrames into a single DataFrame
    concat_df = pd.concat(transp_list, ignore_index=False)
    
    pop_labels = concat_df.index.values
    first_pop = [pop.split("_")[0] for pop in pop_labels]
    second_pop = [pop.split("_")[1] for pop in pop_labels]
    concat_df["first_pop"] = first_pop
    concat_df["second_pop"] = second_pop

    concat_df.first_pop = concat_df.first_pop.astype("category")
    concat_df.first_pop = concat_df.first_pop.cat.set_categories(population_sorter)
    concat_df.second_pop = concat_df.second_pop.astype("category")
    concat_df.second_pop = concat_df.second_pop.cat.set_categories(population_sorter)

    # Clean up the DataFrame
    concat_df.sort_values(["first_pop", "second_pop"], inplace=True)
    concat_df.drop(["first_pop", "second_pop"], axis=1, inplace=True)
    concat_df.index.name = "pop_pairs"

    return concat_df

def plot_exp_heatmap(
    input_df,
    start,
    end,
    title=None,
    output="ExP_heatmap",
    output_suffix="png",
    cmap="Blues",
    populations="1000Genomes",
    vertical_line=True,
    cbar_vmin=None,
    cbar_vmax=None,
    ylabel=None,
    xlabel=None,
    cbar_ticks=None,
    display_limit=None,
    display_values="higher"
):
    """
    Generate an ExP heatmap from a pandas DataFrame of pairwise statistics.

    Parameters
    ----------
    input_df : pandas.DataFrame
        Input data containing pairwise statistics or p-values to visualize.
    start : int
        Start genomic position for the x-axis (region to display).
    end : int
        End genomic position for the x-axis (region to display).
    title : str, optional
        Title of the heatmap.
    output : str, optional
        Output filename (default: ExP_heatmap).
    output_suffix : str, optional
        File extension for the output (default: "png").
    cmap : str, optional
        Colormap for the heatmap (default Blues).
    populations : str or iterable, optional
        Population set to use. By default, expects "1000Genomes" (26 populations).
        For custom populations, provide an iterable of names. The input_df
        should contain all pairwise combinations in both directions.
    vertical_line : bool or list, optional
        If True, draws a vertical line at the center of the x-axis.
        If a list of (position, label) tuples is provided, draws multiple vertical lines with labels.
    cbar_vmin, cbar_vmax : float, optional
        Minimum and maximum values for the colorbar.
    ylabel, xlabel : str, optional
        Custom y-axis/x-axis labels.
    cbar_ticks : list, optional
        Custom ticks for the colorbar.
    display_limit : float, optional
        If set, values above or below this threshold are set to zero, depending on `display_values`.
        Useful for highlighting only the most significant results (e.g., top 1000 SNPs).
    display_values : {'higher', 'lower'}, optional
        Determines which values to retain when applying `display_limit`.
        Use "higher" for rank p-values or distances (default), "lower" for classical p-values.

    Returns
    -------
    The Matplotlib axis object and saves the heatmap to file
    """

    # Helper functions to handle cases where start/end positions are not in the data
    def take_closest_start(sorted_list, target):
        """
        Given a sorted list, return the value equal to target if present,
        otherwise return the smallest value in the list that is >= target.
        If target is greater than all elements, raise ValueError.
        """
        if not sorted_list:
            raise ValueError("Input list is empty")
        pos = bisect_left(sorted_list, target)
        if pos < len(sorted_list) and sorted_list[pos] == target:
            return target
        if pos == len(sorted_list):
            raise ValueError("Your 'start' position index was higher than the range of the input data")
        return sorted_list[pos]

    def take_closest_end(sorted_list, target):
        """
        Given a sorted list, return the value equal to target if present,
        otherwise return the largest value in the list that is <= target.
        If target is less than all elements, raise ValueError.
        """
        if not sorted_list:
            raise ValueError("Input list is empty")
        pos = bisect_left(sorted_list, target)
        if pos < len(sorted_list) and sorted_list[pos] == target:
            return target
        if pos == 0:
            raise ValueError("Your 'end' position index was lower than the range of the input data")
        return sorted_list[pos - 1]
    
    input_df = input_df.copy()
    
    # Crop the input_df according to user defined range
    try: # Given values are in the data (column index)
        input_df = input_df.loc[:, start:end]
    except: # Given values are not in the column index, choose the new closest ones
        sorted_columns = sorted(list(input_df.columns))
        new_start = take_closest_start(sorted_columns, start)
        new_end = take_closest_end(sorted_columns, end)
        input_df = input_df.loc[:, new_start:new_end]
        
        print(f"WARNING: Given 'start' and 'end' datapoints are not found in the input data. New closest datapoints were selected.\nstart={new_start}\nend={new_end}")
        
    # Check the input data for number of populations and input_df shape
    if populations == "1000Genomes":
        print("\nInput data: 1000 Genomes Project, phase 3 (26 populations expected)")
        print("Expecting 650 population pairwise comparisons: ", end="")
        if input_df.shape[0] != 650:
            raise ValueError(f"With selected populations='1000Genomes' option, the input_df was expected to have 650 rows, actual shape was: {input_df.shape[0]} rows, {input_df.shape[1]} columns")
        else:
            print("✓\n")

    else:
        expected_pop_pairs = len(populations) * (len(populations) - 1)
        print(f"\n Input data: {len(populations)} custom populations")
        print(f"Expecting {expected_pop_pairs} population pairs: ", end="")
        if input_df.shape[0] == expected_pop_pairs:
            print("✓\n")
        else:
            raise ValueError(f"With selected populations={populations} option, the input_df was expected to have {expected_pop_pairs} rows, actual shape was: {input_df.shape[0]} rows, {input_df.shape[1]} columns")

    # Apply the display_limit if selected
    if display_limit:
        if display_values == "higher":
            print(f"\nDisplaying only values above the given display_limit: {display_limit}")
            input_df[input_df < display_limit] = 0
            
        elif display_values == "lower":
            print(f"\nDisplaying only values below the given display_limit: {display_limit}")
            input_df[input_df > display_limit] = 0
            
        else:
            raise ValueError(f"plot_exp_heatmap() parameter 'display_values' has unknown value '{display_values}'. The only expected options are 'higher' or 'lower'.")
    
    # Custom expheatmap colormap
    if cmap == "expheatmap":
        from matplotlib import cm
        import matplotlib as mpl

        cmap = cm.gist_ncar_r(np.arange(256))
        cmap[0] = [1.0, 1.0, 1.0, 1.0]
        cmap[-1] = [0, 0, 0.302, 1.0]
        cmap = mpl.colors.ListedColormap(cmap, name='expheatmap_cmap', N=cmap.shape[0])

    # Default heatmap from 1000 Genomes Project data
    if populations == "1000Genomes":
        populations = populations_1000genomes
        
        fig, ax = plt.subplots(figsize=(15, 5))
        
        sns.heatmap(
            input_df,
            yticklabels=populations,
            xticklabels=False,
            vmin=1.301 if cbar_vmin==None else cbar_vmin,
            vmax=4.833 if cbar_vmax==None else cbar_vmax,
            ax=ax,
            cmap=cmap,
        )

        ax.set_title(title)            
        ax.set_xlabel(xlabel)
        ax.set_ylabel("population pairings\n\n    AMR   |     EUR    |     EAS    |    SAS     |       AFR          ")

    # Custom heatmap with user-defined populations
    else:
        # Adjust figure size for large population sample-sets
        fig, ax = plt.subplots(figsize=(15 + (input_df.shape[1] // 1000), 5 + (input_df.shape[0] // 900)))
        
        sns.heatmap(
            input_df,
            yticklabels=populations,
            xticklabels=False,
            vmin=cbar_vmin,
            vmax=cbar_vmax,
            cbar_kws={"ticks": cbar_ticks},
            ax=ax,
            cmap=cmap,
        )

        ax.set_title(title)
        ax.set_ylabel(ylabel)
        ax.set_xlabel(xlabel)
        
    # Set the y-axis tics and labels
    y_axis_len = len(populations) * (len(populations) - 1)
    y_labels_pos = list(np.arange(0, y_axis_len, step=(len(populations)-1)))
    y_labels_pos.append(y_axis_len)
    ax.set_yticks(y_labels_pos)
    ax.set_yticks(np.arange(y_labels_pos[0] + ((len(populations)-1) / 2), y_labels_pos[-1], step=(len(populations)-1)), minor=True)
    ax.tick_params(axis="y", which="minor", length=0)
    ax.set_yticklabels(populations, minor=True)

    # Add vertical line if specified
    if vertical_line is True: # Single vertical line in the middle
        middle = int(input_df.shape[1] / 2)
        ax.axvline(x=middle, linewidth=1, color="grey")
        
    elif vertical_line and hasattr(vertical_line, '__iter__') and not isinstance(vertical_line, str): # Multiple vertical lines with labels
        list_of_columns = input_df.columns.to_list()
        positions_indices = []
        labels = []
        
        for item in vertical_line:
            if not isinstance(item, (tuple, list)) or len(item) != 2:
                raise ValueError(f"Each vertical line item must be a (position, label) tuple, got: {item}")
            
            pos, label = item
            try:
                col_index = list_of_columns.index(pos)
                positions_indices.append(col_index)
                labels.append(label)
                ax.axvline(x=col_index, linewidth=1, color="grey")
            except ValueError:
                print(f"Warning: Position {pos} not found in data columns, skipping this vertical line")
                continue
        
        # Set x-ticks and labels for the valid positions
        if positions_indices:
            ax.set_xticks(positions_indices)
            ax.set_xticklabels(labels)
    
    ax.figure.savefig(f"{output}.{output_suffix}", dpi=400, bbox_inches="tight")
    print(f"ExP heatmap saved into {output}.{output_suffix}")
    return ax
    

    
def prepare_cbar_params(data_df, n_cbar_ticks=4):
    """
    Calculate optimal colorbar parameters for heatmap visualization.
    
    This function analyzes the data range and automatically determines appropriate
    minimum and maximum values for the colorbar, along with evenly spaced tick marks.
    The colorbar bounds are intelligently chosen to provide good visual contrast
    while encompassing the full data range.
    
    Parameters
    ----------
    data_df : pandas.DataFrame
        Input DataFrame containing the data values to be visualized in the heatmap.
        All numeric values in the DataFrame are considered for range calculation.
    n_cbar_ticks : int, optional
        Number of tick marks to display on the colorbar (default: 4).
        Must be at least 2 to show minimum and maximum values.
    
    Returns
    -------
    tuple of (float, float, list)
        A tuple containing:
        - cbar_vmin (float): Minimum value for the colorbar scale
        - cbar_vmax (float): Maximum value for the colorbar scale  
        - cbar_ticks (list): List of evenly spaced tick values for the colorbar
   
    """
    from math import floor
    
    cbar_min = 0
    cbar_max = 0
    data_min = data_df.min().min()
    data_max = data_df.max().max()
    
    # Determine the colorbar minimum and maximum bounds
    if data_min < 0.5:
        cbar_min = 0    
    elif data_min < 1:
        cbar_min = 0.5
    else:
        cbar_min = floor(data_min)
        
    if data_max < 1:
        cbar_max = 1
    else:
        cbar_max = floor(data_max + 1)
        
    # Generate evenly spaced tick marks for the colorbar
    cbar_ticks = np.arange(cbar_min, cbar_max + 0.001, step=(cbar_max - cbar_min)/(n_cbar_ticks-1))
    
    return cbar_min, cbar_max, list(cbar_ticks)



def plot(input_dir, start, end, title, output="ExP_heatmap", cmap="Blues"):
    """
    Generate and save an ExP heatmap from XP-EHH analysis results.

    This function serves as a high-level wrapper that processes a directory of 
    XP-EHH (Cross-Population Extended Haplotype Homozygosity) results and creates 
    a publication-ready heatmap visualization. It combines data loading, processing, 
    and plotting into a single convenient function call.

    Parameters
    ----------
    xpehh_dir : str
        Path to directory containing XP-EHH results as .tsv files.
        Files should be named in the format 'POP1_POP2.*.tsv' where POP1 and POP2 
        are population identifiers from the 1000 Genomes Project.
    start : int
        Start genomic position (inclusive) for the region to visualize.
        Positions are typically in base pairs along a chromosome.
    end : int
        End genomic position (inclusive) for the region to visualize.
        Must be greater than start position.
    title : str
        Main title for the heatmap plot.
    output : str
        Output filename (without extension) for saving the heatmap.
        The plot will be saved as '{output}.png' by default.
    cmap : str, optional
        Matplotlib colormap name for the heatmap visualization (default: "Blues").
        Can also use "expheatmap" for a custom colormap optimized for this analysis.

    Returns
    -------
    None
        The function saves the heatmap to disk but does not return any value.
        Output file is saved as '{output}.png' with 400 DPI resolution.

    Notes
    -----
    This function assumes 1000 Genomes Project population structure and expects
    650 pairwise population comparisons. The function will automatically handle
    cases where the exact start/end positions are not present in the data by
    selecting the closest available positions.
    """
    try:
        plot_input = create_plot_input(input_dir, start=start, end=end)
        
        plot_exp_heatmap(
            plot_input, start=plot_input.columns[0],
            end=plot_input.columns[-1],
            title=title,
            cmap=cmap,
            output=output,
            xlabel="{:,} - {:,}".format(start, end)
        )
    except KeyboardInterrupt:
        print("")
        sys.exit(1)



from exp_heatmap import tests, utils
import sys

def compute(zarr_dir: str, panel_file: str, output_dir: str, test: str, chunked: bool = False):
    utils.check_path_or_exit(zarr_dir)
    utils.check_path_or_exit(panel_file)
    try:
        tests.run(zarr_dir=zarr_dir, panel_file=panel_file, output_dir=output_dir, test=test, chunked=chunked)
    except KeyboardInterrupt:
        print("")
        sys.exit(1)

    print()
    print(f"ZARR dir: {zarr_dir}")
    print(f"Panel file: {panel_file}")
    print(f"Output dir: {output_dir}")



"""
General utilities for file path handling and name extraction.
"""

import re
import os
import sys


def check_path_or_exit(path: str):
    """
    Checks if the given path exists and is accessible.
    
    Prints an error message and exits the program with status code 1
    if the path does not exist or is not accessible.
    
    Args:
        path (str): The file or directory path to check
    """
    if not os.path.exists(path):
        print("Path {} does not exist or it is unaccessible".format(path))
        sys.exit(1)


def name_from_path(path: str):
    """
    Extracts a clean name from a file path by removing trailing slashes and common extensions.
    
    Args:
        path (str): The file path to process
        
    Returns:
        str: The cleaned name without directory path and extensions
    """
    path = re.sub(r"\/+$", "", path)
    path = os.path.basename(path)
    path = re.sub(r"\.vcf$", "", path)
    path = re.sub(r"\.gz$", "", path)
    path = re.sub(r"\.zarr$", "", path)

    return path



import sys
import allel
import zarr

from exp_heatmap import utils


def prepare(recode_file: str, zarr_dir: str) -> None:
    """
    Convert VCF file to ZARR array.
    
    Requirements:
        - zarr version < 3.0.0
    
    Args:
        - recode_file: Path to the input VCF file (recoded, SNPs only)
        - zarr_dir: Path where the ZARR directory will be created
        
    Raises:
        - SystemExit: If zarr version >= 3.0.0 is detected, if input file doesn't exist, or if conversion fails
    """
    # Check zarr version compatibility
    zarr_version = zarr.__version__
    if int(zarr_version.split('.')[0]) >= 3:
        print(f"Error: Unsupported zarr version: {zarr_version}")
        print("Please downgrade to zarr version < 3.0.0:")
        print()
        print("  pip install 'zarr<3.0.0'")
        print()
        sys.exit(1)
    
    # Check if input file exists
    utils.check_path_or_exit(recode_file)

    # Convert VCF file to ZARR array
    try:
        allel.vcf_to_zarr(recode_file, zarr_dir, fields="*", log=sys.stdout)
    except KeyboardInterrupt:
        print("")
        sys.exit(1)
    except Exception as e:
        print(f"Error converting VCF to ZARR: {e}")
        sys.exit(1)
    
    print()
    print(f"Recoded VCF: {recode_file}")
    print(f"ZARR dir: {zarr_dir}")



import allel
import zarr
import os
import numpy as np
import pandas as pd
import sys

from exp_heatmap import xp_utils, utils, rank_tools


def run(
    zarr_dir: str,
    panel_file: str,
    output_dir: str,
    test="xpehh",
    d_tajima_d_size: int = 13,
    chunked: bool = False,
):
    """
    Computes selection tests for all population pairs.

    Args:
        zarr_dir (str): Path to VCF converted to zarr format
        panel_file (str): Path to file defining populations/super-populations  
        output_dir (str): Directory where final *.tsv files will be saved
        test (str): Type of test to compute. Options: 'xpehh', 'xpnsl', 
                   'delta_tajima_d', 'hudson_fst'. Default is 'xpehh'.
        d_tajima_d_size (int): Window size for Delta Tajima's D (number of SNPs).
                              Values between 10-20 work well. Increase if getting
                              empty results. Default is 13.
    """
    print(f"Loading panel file: {panel_file}")
    panel = pd.read_csv(panel_file, sep="\t", usecols=["sample", "pop", "super_pop"])
    pop_pairs = xp_utils.create_pop_pairs(panel)

    print(f"Loading zarr data: {zarr_dir}")
    callset = zarr.open_group(zarr_dir, mode="r")

    gt, positions = xp_utils.filter_by_AF(callset, 0.05, chunked)
    samples = callset["samples"][:]
    
    # Check the sample order
    xp_utils.check_sample_order(samples, panel["sample"])

    name = utils.name_from_path(zarr_dir)

    # For delta Tajima's D, adjust positions for non-overlapping windows
    if test == "delta_tajima_d":
        df = pd.DataFrame({"variant_pos": positions[0::d_tajima_d_size][:-1]})

    else:
        df = pd.DataFrame({"variant_pos": positions})

    df.insert(0, "name", name)

    results = []
    masks = []

    for pair in pop_pairs:
        # Prepare data structures based on test type
        if test in ["xpehh", "xpnsl"]:
            array_pop1 = xp_utils.get_haplotypes(gt, panel, pair[0])
            array_pop2 = xp_utils.get_haplotypes(gt, panel, pair[1])
        elif test in ["delta_tajima_d", "hudson_fst"]:
            array_pop1 = xp_utils.get_pop_allele_counts(gt, panel, pair[0])
            array_pop2 = xp_utils.get_pop_allele_counts(gt, panel, pair[1])

        print(f"Computing {test.upper()} for pair {pair[0]} vs {pair[1]}")
        print(f"Population {pair[0]} dimensions: {' '.join(map(str, array_pop1.shape))}")
        print(f"Population {pair[1]} dimensions: {' '.join(map(str, array_pop2.shape))}")
        print(f"Number of positions: {len(positions)}")

        # Compute the selected test
        if test == "xpehh":
            result = allel.xpehh(
                h1=array_pop1,
                h2=array_pop2,
                pos=positions,
                map_pos=None,
                min_ehh=0.05,
                include_edges=False,
                gap_scale=20000,
                max_gap=200000,
                is_accessible=None,
                use_threads=True,
            )
        elif test == "xpnsl":
            result = allel.xpnsl(
                h1=array_pop1,
                h2=array_pop2,
                use_threads=True,
            )
        elif test == "delta_tajima_d":
            result = allel.moving_delta_tajima_d(
                ac1=array_pop1,
                ac2=array_pop2,
                size=d_tajima_d_size,
                start=0,
                stop=None,
                step=d_tajima_d_size,
            )
        elif test == "hudson_fst":
            num, den = allel.hudson_fst(
                ac1=array_pop1,
                ac2=array_pop2,
            )
            result = num / den

        # Create mask for NaN values
        mask = np.isnan(result)
        results.append(result)
        masks.append(mask)

    # Create combined NaN mask - True where any population pair has NaN
    combined_nan_mask = np.logical_or.reduce(masks)
    # Convert to boolean mask for valid positions (not NaN)
    valid_positions_mask = ~combined_nan_mask
    
    num_masked = np.sum(combined_nan_mask)
    print("Applying NaN mask for all results")
    print("Number of results removed from each file: {}".format(num_masked))

    # Check if all positions are masked
    if num_masked == len(valid_positions_mask):
        print("\n=== WARNING ===")
        print("All positions have NaN results. No output will be generated.")
        if test == "delta_tajima_d":
            print("For delta Tajima's D, try increasing the 'd_tajima_d_size' parameter.")
        return

    os.makedirs(output_dir, exist_ok=True)

    for pair, result_data in zip(pop_pairs, results):
        result_path = os.path.join(output_dir, "_".join(pair) + ".tsv")

        # add results to the dataframe with coordinates
        df[test] = result_data

        # Compute ascending and descending log10 rank p-values
        for ascending in [True, False]:
            df.sort_values(by=test, inplace=True, ascending=ascending)
            test_results = df[test].values
            ranks = rank_tools.compute_ranks(test_results)
            rank_p_vals = rank_tools.compute_rank_p_vals(ranks)
            log_10_p_vals = rank_tools.compute_log_10_p_vals(rank_p_vals)

            suffix = "ascending" if ascending else "descending"
            df[f"-log10_p_value_{suffix}"] = log_10_p_vals

        df.sort_values(by="variant_pos", inplace=True, ascending=True)

        # Save only positions without NaN values
        df[valid_positions_mask].to_csv(result_path, index=False, sep="\t")

        print(f"Results saved to: {result_path}")



"""
Specific utilities for genetic data analysis and population genetics calculations.
"""

import numpy as np
import allel
import sys
import pandas as pd
from itertools import combinations
from typing import List, Tuple, Any

def create_pop_pairs(panel: pd.DataFrame) -> List[Tuple[str, str]]:
    """
    Create all unique population pairs from a population panel.
    
    Args:
        panel: DataFrame with population information, must contain 'pop' column
        
    Returns:
        List of tuples containing population pair combinations
        
    Example:
        >>> panel = pd.DataFrame({'pop': ['AFR', 'EUR', 'ASN']})
        >>> create_pop_pairs(panel)
        [('AFR', 'EUR'), ('AFR', 'ASN'), ('EUR', 'ASN')]
    """
    populations = np.unique(panel["pop"].values)
    return list(combinations(populations, 2))

def get_haplotypes(gt_array: allel.GenotypeArray, panel: pd.DataFrame, pop: str) -> allel.HaplotypeArray:
    """
    Extract haplotype data for a specific population.
    
    Args:
        gt_array: Genotype array containing all samples
        panel: Population panel DataFrame with sample metadata
        pop: Population identifier to extract
        
    Returns:
        Haplotype array for the specified population
    """
    # Get indices of samples belonging to the specified population
    indices_pop = panel.index[panel["pop"] == pop]

    # Extract genotype data for the population
    gt_pop = gt_array.take(indices_pop, axis=1)

    return gt_pop.to_haplotypes()



def get_pop_allele_counts(gt: allel.GenotypeArray, panel: pd.DataFrame, pop: str) -> allel.AlleleCountsArray:
    """
    Calculate allele counts for a specific population.
    
    Args:
        gt: Genotype array containing all samples
        panel: Population panel DataFrame with sample metadata
        pop: Population identifier to calculate counts for
        
    Returns:
        Allele counts array for the specified population
    """
    # Get indices of samples belonging to the specified population
    indices_pop = panel.index[panel["pop"] == pop]

    # Extract genotype data for the population
    gt_pop = gt.take(indices_pop, axis=1)

    # Calculate allele counts for the population (input for PBS analysis)
    ac = gt_pop.count_alleles()
    return ac

def filter_by_AF(callset: Any, af_threshold: float, chunked: bool = False) -> Tuple[allel.GenotypeChunkedArray, allel.SortedIndex]:
    """
    Filter variants by alternate allele frequency threshold.
    
    Args:
        callset: Zarr callset containing variant data
        af_threshold: Minimum alternate allele frequency threshold
        
    Returns:
        Tuple of (filtered_genotypes, filtered_positions) where variants
        have alternate allele frequency > af_threshold
    """
    af = callset["variants/AF"][:]
    loc_variant_selection = af[:, 0] > af_threshold
    gt_zarr = callset["calldata/GT"]

    # Load the genotype as chunked array for memory efficiency
    if chunked:
        print("Using chunked array to avoid memory exhaustion")
        gt = allel.GenotypeChunkedArray(gt_zarr)
    else:
        print("Attempting to load full array into memory")
        gt = allel.GenotypeArray(gt_zarr)

    # Filter variants above the threshold
    gt_filtered = gt.compress(loc_variant_selection, axis=0)
    
    # Filter positions correspondingly
    positions = allel.SortedIndex(callset["variants/POS"])
    positions_filtered = positions.compress(loc_variant_selection, axis=0)

    return gt_filtered, positions_filtered

def check_sample_order(zarr_samples, panel_samples) -> None:
    """
    Checks the sample order between zarr and panel data.
    Exits the program if samples don't match or are in different order.
    
    Args:
        zarr_samples: Array or list of sample names from zarr file
        panel_samples: Array or pandas Series of sample names from panel file
    """
    # Convert to lists for easier manipulation
    zarr_samples_list = zarr_samples.tolist() if hasattr(zarr_samples, 'tolist') else list(zarr_samples)
    panel_samples_list = panel_samples.tolist() if hasattr(panel_samples, 'tolist') else list(panel_samples)
    
    print(f"Zarr samples: {len(zarr_samples_list)}")
    print(f"Panel samples: {len(panel_samples_list)}")
    
    # Verify that both datasets have the same number of samples
    if len(zarr_samples_list) != len(panel_samples_list):
        print(f"\nERROR: Different number of samples!")
        print(f"Zarr has {len(zarr_samples_list)} samples, panel has {len(panel_samples_list)} samples")
        sys.exit(1)
    
    # Check if samples are the same (regardless of order)
    zarr_set = set(zarr_samples_list)
    panel_set = set(panel_samples_list)
    
    if zarr_set != panel_set:
        missing_in_panel = zarr_set - panel_set
        missing_in_zarr = panel_set - zarr_set
        
        print(f"\nERROR: Different samples!")
        if missing_in_panel:
            print(f"Samples in zarr but not in panel: {missing_in_panel}")
        if missing_in_zarr:
            print(f"Samples in panel but not in zarr: {missing_in_zarr}")
        sys.exit(1)
    
    # Verify that sample order matches between datasets
    order_matches = True
    mismatches = []
    
    for i, (zarr_sample, panel_sample) in enumerate(zip(zarr_samples_list, panel_samples_list)):
        if zarr_sample != panel_sample:
            order_matches = False
            mismatches.append((i, zarr_sample, panel_sample))
    
    if order_matches:
        print(f"\nSUCCESS: Sample order is identical!")
        print(f"All {len(zarr_samples_list)} samples are in the same order.")
    else:
        print(f"\nWARNING: Sample order differs!")
        print(f"Found {len(mismatches)} mismatches:")
        
        # Display the first 10 mismatched positions for debugging
        for i, zarr_sample, panel_sample in mismatches[:10]:
            print(f"  Position {i+1}: Zarr='{zarr_sample}' vs Panel='{panel_sample}'")
        
        if len(mismatches) > 10:
            print(f"  ... and {len(mismatches) - 10} more mismatches")
        
        print(f"\nFirst 10 samples comparison:")
        print("Position\tZarr\t\tPanel")
        for i in range(min(10, len(zarr_samples_list))):
            match_status = "✓" if zarr_samples_list[i] == panel_samples_list[i] else "✗"
            print(f"{i+1}\t{zarr_samples_list[i]}\t{panel_samples_list[i]}\t{match_status}")
        
        print("\nOrder of samples in panel file does not match order of samples in zarr.")
        print("It is possible that you are using wrong panel file path e.g. from different phase than your variant data.")
        sys.exit(1)